#include <gb/gb.h>
#include <stdio.h>
#include <ctype.h>
#include "game.h"
#include "title.h"
#include "main_tileset.h"
#include "main_sprites.h"

// Background scrollable tilemap
const unsigned char back_tiles[] = {
	0, 0, 0, 1, 0, 0, 2, 0, 0, 3, 0, 0, 2, 0, 0, 1, 0, 0, 0, 0,
	0, 0, 0, 1, 0, 0, 2, 0, 0, 3, 0, 0, 2, 0, 0, 1, 0, 0, 0, 0,
	0, 0, 0, 1, 0, 0, 2, 0, 0, 3, 0, 0, 2, 0, 0, 1, 0, 0, 0, 0,
	0, 0, 0, 4, 7, 7, 5, 7, 7, 6, 7, 7, 5, 7, 7, 4, 0, 0, 0, 0,
};

// Random tables
const unsigned char random_64[256] = {
	41, 63, 13, 33, 22, 21, 49, 4,
	9, 45, 30, 33, 3, 15, 9, 47,
	44, 18, 59, 31, 8, 0, 45, 25,
	40, 53, 22, 9, 29, 8, 1, 26,
	50, 45, 34, 1, 27, 26, 42, 62,
	40, 41, 25, 31, 62, 62, 20, 5,
	8, 41, 55, 42, 42, 14, 44, 35,
	54, 3, 61, 35, 35, 57, 38, 1,
	59, 31, 47, 59, 19, 3, 56, 58,
	0, 58, 22, 24, 8, 3, 29, 31,
	23, 39, 62, 19, 51, 35, 20, 51,
	50, 9, 57, 61, 43, 36, 30, 48,
	42, 27, 27, 21, 40, 24, 3, 0,
	58, 41, 60, 20, 8, 38, 13, 5,
	1, 20, 57, 34, 13, 9, 12, 18,
	3, 17, 51, 30, 29, 17, 16, 47,
	41, 42, 14, 25, 1, 32, 50, 25,
	60, 60, 51, 27, 38, 46, 12, 0,
	46, 55, 15, 29, 31, 6, 60, 7,
	17, 33, 18, 14, 33, 7, 37, 7,
	30, 20, 49, 15, 5, 36, 46, 12,
	48, 59, 37, 54, 10, 37, 1, 22,
	6, 25, 41, 14, 17, 35, 13, 26,
	5, 3, 63, 21, 31, 16, 5, 63,
	51, 32, 44, 61, 30, 44, 31, 33,
	61, 48, 55, 44, 28, 17, 21, 17,
	49, 2, 46, 60, 11, 40, 39, 19,
	5, 60, 34, 60, 63, 50, 23, 31,
	14, 24, 38, 45, 51, 0, 43, 33,
	10, 33, 20, 6, 4, 59, 47, 57,
	23, 58, 7, 48, 52, 29, 49, 59,
	58, 23, 34, 10, 41, 54, 4, 48,
};

const unsigned char random_5[256] = {
	4, 4, 4, 2, 1, 4, 3, 0,
	3, 2, 0, 2, 4, 2, 2, 2,
	4, 4, 4, 0, 4, 0, 2, 4,
	3, 3, 2, 0, 4, 4, 4, 1,
	3, 0, 0, 4, 1, 4, 1, 0,
	3, 0, 4, 2, 4, 1, 3, 0,
	2, 2, 3, 4, 3, 3, 4, 4,
	4, 0, 0, 2, 2, 2, 2, 4,
	3, 3, 4, 4, 3, 0, 0, 2,
	4, 4, 3, 1, 4, 0, 2, 3,
	0, 2, 1, 4, 0, 3, 1, 1,
	2, 0, 2, 4, 3, 3, 0, 4,
	0, 0, 4, 1, 3, 3, 1, 2,
	1, 2, 0, 3, 2, 4, 2, 1,
	1, 3, 0, 1, 2, 0, 3, 4,
	0, 3, 1, 0, 4, 2, 0, 0,
	2, 4, 0, 3, 2, 1, 0, 1,
	4, 0, 1, 4, 0, 2, 2, 4,
	0, 4, 3, 4, 3, 1, 0, 3,
	3, 0, 2, 3, 0, 3, 3, 1,
	3, 1, 2, 1, 0, 3, 4, 3,
	3, 4, 1, 0, 1, 4, 0, 4,
	0, 1, 3, 3, 4, 2, 1, 1,
	2, 3, 4, 2, 2, 4, 0, 2,
	4, 2, 1, 1, 3, 0, 0, 2,
	4, 4, 0, 3, 4, 2, 2, 3,
	4, 3, 4, 1, 2, 2, 3, 3,
	4, 1, 2, 4, 4, 1, 4, 3,
	0, 2, 3, 2, 1, 3, 3, 0,
	1, 4, 2, 1, 2, 2, 2, 1,
	2, 2, 1, 2, 1, 2, 3, 0,
	3, 3, 1, 1, 0, 4, 2, 0,
};

const UINT16 bonus_freq_table[] = {
	1, 0x590, 0x650, 0
};

// [2] = {L, R}, [9] diff�rence entre le bruit et le vaisseau (0=le bruit est � gauche, 4=m�me endroit, 8=le bruit est � droite)
const unsigned char volumeTable[2][9] = {
	{15, 15, 15, 15,
	 15, 7, 4, 2, 0},

	{0, 2, 4, 7, 15,
	 15, 15, 15, 15},
};

const unsigned char palette_table[4] = {
	0xE4,
	0x90,
	0x40,
	0x00,
};

// Positions on screen, for animation
int currentShipPosX = 27 + 2 * 24, currentShipPosY = 16 + 150;
UINT8 random_counter;
unsigned char score[5];
UINT16 *current_sound_freq_ptr;
UINT8 current_sound_freq_divider;

// Setup for

void sound_setup() {
	UINT8 *addr = 0xFF30, i;
	NR50_REG = 0xff;
	NR51_REG = BITS(1, 1, 0, 1, 1, 1, 1, 0);
	NR52_REG = 0x80;
	for (i = 0; i < 16; i++)
		*addr++ = 0xf0;
}

/**
 * Version without sweep (volume nor frequency). No end of sound.
 * @param patternDuty 0..3 (12.5%, 25%, 50%, 75%)
 * @param volume 0..15
 * @param frequency wave frequency
 */
void sq1_setup(UINT8 patternDuty, UINT8 volume, UINT16 frequency) {
	const UINT8 soundLength = 63;
	NR10_REG = 0;
	// FIXME: should be << 6, but due to a compiler bug, 5
	NR11_REG = patternDuty << 5 | soundLength;
	NR12_REG = volume << 4 | 0 << 3 | 0;
	NR13_REG = frequency & 0xff;
	NR14_REG = 1 << 7 | 0 << 6 | (frequency >> 8 & 0x7);
}

/**
 * Same as sq1_setup.
 * @param patternDuty 0..3 (12.5%, 25%, 50%, 75%)
 * @param volume 0..15
 * @param frequency wave frequency
 */
void sq2_setup(UINT8 patternDuty, UINT8 volume, UINT16 frequency) {
	const UINT8 soundLength = 63;
	// FIXME: should be << 6, but due to a compiler bug, 5
	NR21_REG = patternDuty << 5 | soundLength;
	NR22_REG = volume << 4 | 0 << 3 | 0;
	NR23_REG = frequency & 0xff;
	NR24_REG = 1 << 7 | 0 << 6 | (frequency >> 8 & 0x7);
}

/**
 * Same as sq1_setup.
 * @param frequency wave frequency
 */
void sq3_setup(UINT16 frequency) {
	const UINT8 soundLength = 255;
	if (frequency == 0)
		NR30_REG = 0x00;
	else
		NR30_REG = 0x80;
	NR31_REG = soundLength;
	NR32_REG = 1 << 5;			// hard coded volume because of bugs in the compiler
	NR33_REG = frequency & 0xff;
	NR34_REG = 1 << 7 | 0 << 6 | (frequency >> 8 & 0x7);
}

/**
 * Play a square sound with an ennemy at a given distance.
 * @param pan distance between the object and the player, 0..8. 0 is played all the left, 8 all the right.
 * @param dist distance, multiplier (0..64..95 -> far..near..away)
 */
void sound_set(UINT8 duty, UINT8 pan, UINT8 dist) {
	UINT8 volL, volR;
	if (pan < 0)
		pan = 0;
	if (pan > 8)
		pan = 8;
	volL = volumeTable[0][pan];
	volR = volumeTable[1][pan];
	sq1_setup(duty, volL, 0x640 + dist);
	sq2_setup(duty, volR, 0x640 + dist);
}

/**
 * Initialize the sprite system and create the default sprites.
 * Sprite allocation:
 * 0-3: main aircraft
 * 4-7: ennemy
 * 30-39: score
 */
void sprites_init() {
	UINT8 i;
	set_sprite_data(0, 25, main_sprites);
	// Aircraft
	set_sprite_tile(0, 1);
	set_sprite_prop(0, 0x10);
	set_sprite_tile(1, 3);
	set_sprite_prop(1, 0x10);
	set_sprite_tile(2, 2);
	set_sprite_prop(2, 0x10);
	set_sprite_tile(3, 4);
	set_sprite_prop(3, 0x10);
	// Ennemy
	set_sprite_tile(4, 15);
	set_sprite_tile(5, 16);
	set_sprite_tile(6, 17);
	set_sprite_tile(7, 18);

	for (i = 0; i < 10; i++) {
		move_sprite(30 + i, 156, 144 - 16 - 8 - i * 8);
		//		move_sprite(39 - i, 48 + i * 8, 16);
		set_sprite_tile(30 + i, 5);
		set_sprite_prop(30 + i, 0x10);
	}
}

/**
 * Position the ship on a horizontal row. Does an automatic animation to place the ship.
 * @param row 0..4
 */
void position_ship(UINT8 row) {
	UINT8 destPosX = 27 + row * 24;
	const UINT8 destPosY = 16 + 120;
	if (destPosX > currentShipPosX)
		currentShipPosX += 3;
	else if (destPosX < currentShipPosX)
		currentShipPosX -= 3;
	if (destPosY < currentShipPosY)
		currentShipPosY--;
	move_sprite(0, currentShipPosX, currentShipPosY);
	move_sprite(1, currentShipPosX + 8, currentShipPosY);
	move_sprite(2, currentShipPosX, currentShipPosY + 8);
	move_sprite(3, currentShipPosX + 8, currentShipPosY + 8);
}

/**
 * Position the ennemy ship.
 * @param distanceFromShip where to display the ship
 */
void position_ennemy(UINT8 row, UINT8 distanceFromShip) {
	UINT8 x = 27 + row * 24;
	UINT8 y;
	if (distanceFromShip >= 120 + 32 + 16)
		y = 0;
	else
		y = 16 + 120 + 32 - distanceFromShip;
	move_sprite(4, x, y);
	move_sprite(5, x + 8, y);
	move_sprite(6, x, y + 8);
	move_sprite(7, x + 8, y + 8);
}

void draw_score() {
	set_sprite_tile(30, 5 + (score[0] & 0xf));
	set_sprite_tile(31, 5 + (score[0] >> 4));
	set_sprite_tile(32, 5 + (score[1] & 0xf));
	set_sprite_tile(33, 5 + (score[1] >> 4));
	set_sprite_tile(34, 5 + (score[2] & 0xf));
	set_sprite_tile(35, 5 + (score[2] >> 4));
	set_sprite_tile(36, 5 + (score[3] & 0xf));
	set_sprite_tile(37, 5 + (score[3] >> 4));
	if (score[4] > 0) {
		set_sprite_tile(38, 5 + (score[4] & 0xf));
		set_sprite_tile(39, 5 + (score[4] >> 4));
	} else {
		set_sprite_tile(38, 24);
		set_sprite_tile(39, 23);
	}
}

void game() {
	int i;
	UINT8 lastKeys = joypad();
	fixed scrollingSpeed;
	UINT16 scrollFixedPoint, scrollingValue;
	UINT16 currentEnnemyDistance;
	UINT8 currentShipRow, currentEnnemyRow, nextIsEnnemy;
	UINT8 lostCounter;
	UINT8 currentPalette, frameCounterForPalette = 0;
	UINT8 pause;
	UINT16 null = 0;

	scrollFixedPoint = 0;
	scrollingValue = 0;
	currentShipRow = 2, currentEnnemyRow = 0, nextIsEnnemy = 0;
	scrollingSpeed.w = 0x1000;
	currentEnnemyDistance = 200;
	lostCounter = 0;
	score[0] = score[1] = score[2] = score[3] = score[4] = 0;
	currentPalette = 0;
	pause = 0;

	DISPLAY_OFF;
	sound_setup();
	// Build initial background
	set_bkg_data(0, 13, main_tileset);
	for (i = 0; i < 32; i += 4)
		set_bkg_tiles(0, i, 20, 8, back_tiles);
	// Sprites
	sprites_init();

	LCDC_REG = BITS(1, 0, 0, 0, 0, 0, 1, 1);
	OBP1_REG = 0xE4;
	DISPLAY_ON;
	
	current_sound_freq_ptr = &null;
	
	// Game
	while (1) {
		UINT8 keys = joypad();
		UINT8 keysPressed = keys & ~lastKeys;
		UINT8 bg_invisible = (score[1] >> 4) >= 2 && lostCounter == 0;
		// Moving
		if (keysPressed & J_RIGHT && currentShipRow < 4)
			currentShipRow++;
		if (keysPressed & J_LEFT && currentShipRow > 0)
			currentShipRow--;
		if (keysPressed & J_START) {
			pause = !pause;
			sq1_setup(0, 0, 0);
			sq2_setup(0, 0, 0);
			sq3_setup(0);
		}
		lastKeys = keys;

		if (pause)
			continue;

		// Scrolling
		scrollFixedPoint += scrollingSpeed.w >> 5;
		if (scrollFixedPoint >= 256) {
			// Various events on scroll
			UINT8 scrollQty = scrollFixedPoint >> 8;
			scrollingValue += scrollQty;
			currentEnnemyDistance -= scrollQty;
			scroll_bkg(0, -scrollQty);
			if (lostCounter == 0)
				increment_score(scrollQty + scrollQty + scrollQty);
			scrollFixedPoint &= 255;
		}
		// Increment scrolling speed with level
		if (scrollingSpeed.b.h < 0xf0)
			scrollingSpeed.w += 0x4;
		if (currentEnnemyDistance < 38 && currentEnnemyRow == currentShipRow) {
			if (nextIsEnnemy) {
				lostCounter = 1;
				// Aircraft becomes exploded
				set_sprite_tile(0, 19);
				set_sprite_tile(1, 20);
				set_sprite_tile(2, 21);
				set_sprite_tile(3, 22);
			} else {
				// Was a bonus
				currentEnnemyDistance = 15;
				increment_score(0x1500);
				current_sound_freq_ptr = bonus_freq_table;
				current_sound_freq_divider = 7;
			}
		}
		// Ennemy at bottom -> regenerate
		if (currentEnnemyDistance < 16) {
			currentEnnemyDistance = 160 + random_64[random_counter];
			currentEnnemyRow = random_5[random_counter];
			nextIsEnnemy = (random_64[random_counter] & 15) >= 12;
			random_counter++;
		}
		// Lose "animation"
		if (lostCounter > 0) {
			lostCounter++;
			if (lostCounter > 40)
				break;
		}

		// Fade out animation
		frameCounterForPalette++;
		if (frameCounterForPalette == 18) {
			frameCounterForPalette = 0;
			if (bg_invisible) {
				// Clear
				if (currentPalette < 3)
					currentPalette++;
			} else {
				// Normal
				if (currentPalette > 0)
					currentPalette--;
			}
		}
		
		// Current sound (sq3)
		if (current_sound_freq_ptr[0] != 0) {
			current_sound_freq_divider++;
			if (current_sound_freq_divider == 8) {
				current_sound_freq_divider = 0;
				current_sound_freq_ptr++;
				sq3_setup(current_sound_freq_ptr[0]);
			}
		}

		wait_vbl_done();
		// Setup gfx at VBLANK
		OBP0_REG = BGP_REG = palette_table[currentPalette];
		position_ship(currentShipRow);
		if (currentEnnemyDistance >= 120 + 32 + 16) {
			position_ennemy(0, 0);
			sq1_setup(0, 0, 0);
			sq2_setup(0, 0, 0);
		} else {
			UINT8 duty = 2;
			if (nextIsEnnemy)
				duty = 0;
			position_ennemy(currentEnnemyRow, currentEnnemyDistance);
			sound_set(duty, 4 + (currentEnnemyRow - currentShipRow), currentEnnemyDistance);
		}
		draw_score();
	}

	// When game over happens
	OBP0_REG = BGP_REG = 0xE4;
	sq1_setup(0, 0, 0);
	sq2_setup(0, 0, 0);
	sq3_setup(0);
	while (1) {
		UINT8 keys = joypad();
		UINT8 keysPressed = keys & ~lastKeys;
		lastKeys = keys;
		if (keysPressed & J_START)
			break;
		wait_vbl_done();
	}
}

void main() {
	sound_setup();
	sq1_setup(0, 0, 0);
	sq2_setup(0, 0, 0);
	sq3_setup(0);
	while (1) {
		title();
		game();
	}
}
